'''
curate.py

Curates the InifiniteMario data, working with the values defined in
hierarchy.py.

Copyright 2010 Peter Mawhorter
This software is distributed under the terms of the GNU General Public License.
'''

import ore.chunk
import ore.position

import hierarchy

def load(filename):
  with open(filename, 'rU') as fin:
    data = fin.read()
  chunks = data.split("Chunk:\n")[1:]
  lib = []
  for c in chunks:
    lib.append(unpack(c))
  return lib

def save(chunks, filename):
  with open(filename, 'w') as fout:
    for c in chunks:
      fout.write(pack(c))

def unpack(data):
  pos = None
  mode = None
  onex = ore.position.GridPosition(1, 0)
  origin = ore.position.GridPosition(0, 0)
  result = ore.chunk.Chunk()
  for line in data.split('\n'):
    if line == 'Tags:':
      mode = 'tags'
      continue
    if line == 'Origin:':
      mode = 'origin'
      continue
    if line == 'Anchors:':
      mode = 'anchors'
      continue
    if line == 'Map:':
      mode = 'map'
      pos = ore.position.GridPosition(0, 0) - origin
      continue
    elif line == 'Enemies:':
      mode = 'enemy'
      pos = ore.position.GridPosition(0, 0) - origin
      continue
    elif line == 'End':
      mode = None
      continue
    else:
      if mode == 'tags':
        things = line.strip().split()
        if len(things) == 1:
          result.tag(things[0], True)
        elif len(things) == 2:
          tag = things[0]
          val = things[1]
          try:
            val = int(val)
          except:
            try:
              val = float(val)
            except:
              pass
          result.tag(tag, val)
        else:
          raise ValueError("Tags can't have more than a name and a value.")
      elif mode == 'origin':
        first, second = line.split(',')
        origin = ore.position.GridPosition(
          int(first.strip()),
          int(second.strip())
        )
      elif mode == 'anchors':
        first, second = line.split(',')
        anchor = ore.position.GridPosition(
          int(first.strip()),
          int(second.strip())
        )
        result.add_anchor(anchor)
      elif mode in ('map', 'enemy'):
        for c in line:
          if c == '`':
            result.add_anchor(pos)
          elif c != ' ':
            result.add_component(hierarchy.char_to_type[c].at(pos))
          pos += onex
        pos = ore.position.GridPosition(-origin.x, pos.y + 1)
  return result

def mapstr(component):
  if component.real_type in hierarchy.type_to_char\
  and component.is_instance_of(hierarchy.platform):
    return hierarchy.type_to_char[component.real_type]
  return None

def nmestr(component):
  if component.real_type in hierarchy.type_to_char\
  and component.is_instance_of(hierarchy.enemy):
    return hierarchy.type_to_char[component.real_type]
  return None

def pack(chunk):
  minx = None
  maxx = None
  miny = None
  maxy = None
  for pos in chunk.coverage:
    if minx == None or pos.x < minx:
      minx = pos.x
    if maxx == None or pos.x > maxx:
      maxx = pos.x
    if miny == None or pos.y < miny:
      miny = pos.y
    if maxy == None or pos.y > maxy:
      maxy = pos.y

  result = 'Chunk:\nTags:\n'
  for t, v in chunk.tags.items():
    if v == True:
      result += t + '\n'
    else:
      result += "{} {}\n".format(t, v)

  result += 'Origin:\n'
  result += "{}, {}\n".format(-minx, -miny)

  result += 'Anchors:\n'
  for a in chunk.anchors:
    result += "{}, {}\n".format(a.x, a.y)

  result += 'Map:\n'
  for y in range(miny, maxy):
    for x in range(minx, maxx):
      components = chunk.get(ore.position.GridPosition(x, y))
      filled = False
      for c in components:
        if mapstr(c):
          result += mapstr(c)
          filled = True
          break
      if not filled:
        result += ' '
    result += '\n'
  result += 'Enemies:\n'
  for y in range(miny, maxy):
    for x in range(minx, maxx):
      components = chunk.get(ore.position.GridPosition(x, y))
      filled = False
      for c in components:
        if nmestr(c):
          result += nmestr(c)
          filled = True
          break
      if not filled:
        result += ' '
    result += '\n'
  result += 'End\n'

  return result
