'''
chunk.py

Code that defines chunks of components and how they might interact.

Copyright 2010 Peter Mawhorter
This software is distributed under the terms of the GNU General Public License.
'''

import collections

from . import component
from . import position

class Chunk:
  '''
  A chunk is a collection of components. It has a notion of anchor points,
  which can be used to specify where chunks should be placed relative to one
  another. Components are stored as a map from Positions to sets of Component
  objects. Chunks also have a set of tags that can be used for selection.
  '''
  _cmpdict = None
  anchors = None
  tags = None

  @property
  def components(self):
    for pos in self._cmpdict:
      for component in self._cmpdict[pos]:
        yield component

  @property
  def coverage(self):
    for pos in self._cmpdict:
      yield pos

  def __init__(self, components=None, anchors=None, tags=None):
    if isinstance(components, dict):
      self._cmpdict = components
    elif isinstance(components, collections.Iterable):
      for c in components:
        self.add_component(c)
    else:
      self._cmpdict = {}
    self.anchors = anchors or set()
    self.tags = tags or {}

  def __str__(self):
    if self.components:
      return "a chunk with {} anchors near {}".format(
        len(self.anchors),
        next(self.components).position
      )
    else:
      return "an empty chunk with {} anchors".format(len(self.anchors))

  def __repr__(self):
    return "Chunk({}, {}, {})".format(
      repr(self.components),
      repr(self.anchors),
      repr(self.tags)
    )

  def get(self, pos):
    '''
    Gets a set of components at the specified position. Returns an empty set if
    there are no components at that location.
    '''
    if pos in self._cmpdict:
      return self._cmpdict[pos]
    return set()

  def tag(self, tag, value):
    '''
    Adds the given tag with the given value.
    '''
    self.tags[tag] = value

  def untag(self, tag):
    '''
    Removes the given tag.
    '''
    del self.tags[tag]

  def add(
    self,
    other,
    take_relative_to=position.origin,
    put_relative_to=position.origin
  ):
    '''
    Adds the components from the other chunk to this chunk, using the given
    anchor points (or absolute coordinates, if no anchor points are given) to
    align the chunks. Alternatively, if other is just a component, adds it to
    this chunk. Likewise, if other is a Position, it is added to the anchors
    for this chunk.
    '''
    m = None
    if isinstance(other, position.Position):
      m = self.add_anchor
    if isinstance(other, Component):
      m = self.add_component
    elif isinstance(other, Chunk):
      m = self.add_chunk
    else:
      raise TypeError(
        "The 'other' argument must be a Component, Position, or Chunk."
      )
    m(other, myap, otherap)

  def add_anchor(
    self,
    anchor,
    take_relative_to=position.origin,
    put_relative_to=position.origin
  ):
    '''
    Adds an anchor to this chunk, inserting it relative to the given point,
    using it's position relative to the other given reference point.
    '''
    self.anchors.add(put_relative_to + (anchor - take_relative_to))

  def add_component(
    self,
    component,
    take_relative_to=position.origin,
    put_relative_to=position.origin
  ):
    '''
    Adds a component to this chunk, inserting it relative to the given point,
    using it's position relative to the other given reference point.
    '''
    new = component.at(
      put_relative_to + (component.position - take_relative_to)
    )
    if new.position in self._cmpdict:
      self._cmpdict[new.position].add(new)
    else:
      self._cmpdict[new.position] = {new}

  def add_chunk(
    self,
    chunk,
    take_relative_to=position.origin,
    put_relative_to=position.origin
  ):
    '''
    Adds all of the components and anchor points in the given chunk to this
    one. This method throws an error if the chunks aren't compatible given the
    offsets passed in, and it performs component merging, giving preference to
    the components in this chunk.
    '''
    for a in chunk.anchors:
      self.add_anchor(a, take_relative_to, put_relative_to)
    for pos in chunk._cmpdict:
      trans = put_relative_to + (pos - take_relative_to)
      if trans in self._cmpdict:
        targets = list(self._cmpdict[trans])
        for cmp in chunk._cmpdict[pos]:
          dealt_with = False
          for t in targets:
            #print("resolving:", cmp, t)
            if cmp.can_overlap(t):
              #print("overlap allowed...")
              # overlap is fine, but keep checking targets
              continue
            elif cmp.is_satisfied_by(t):
              #print("t subsumes cmp...")
              # t subsumes cmp
              dealt_with = True
              break
            elif t.is_satisfied_by(cmp):
              #print("cmp subsumes t...")
              targets.remove(t)
              targets.append(cmp.at(trans))
              self._cmpdict[trans].remove(t)
              self._cmpdict[trans].add(cmp.at(trans))
              dealt_with = True
              # cmp replaces t (keep checking overlap)
            else:
              raise ValueError(
                "Can't add chunks: incoming component {} is incompatible with "
                "existing component {}.".format(cmp, t)
              )
          if not dealt_with:
            #print("adding component.")
            self._cmpdict[trans].add(cmp.at(trans))
      else:
        self._cmpdict[trans] = set()
        for cmp in chunk._cmpdict[pos]:
          self._cmpdict[trans].add(cmp.at(trans))

  def remove_component(self, target):
    '''
    Removes the given component from this chunk. If it's not actually in this
    chunk, nothing happens. Note that this operation is dangerous: don't
    try it while iterating over the components of this chunk. If it removes
    the component, this function returns True, otherwise it returns False (for
    example if the given component wasn't in this chunk to begin with).
    '''
    for pos in self._cmpdict:
      if target in self._cmpdict[pos]:
        self._cmpdict[pos].remove(target)
        return True
    return False

  def is_compatible_with(
    self,
    other,
    take_relative_to=position.origin,
    put_relative_to=position.origin
  ):
    '''
    Checks chunk compatibility at the given anchor points. Note that chunk
    compatibility isn't quite reflexive, but it's always the case that if
    a.is_compatible_with(b), then a.add(b) is allowed (b.add(a) may or may not
    be).
    '''
    for pos in other._cmpdict:
      trans = put_relative_to + (pos - take_relative_to)
      if trans in self._cmpdict:
        # Make a copy so that we're not changing things:
        targets = list(self._cmpdict[trans])
        for cmp in other._cmpdict[pos]:
          for t in targets:
            if cmp.can_overlap(t):
              # overlap is fine, but keep checking targets
              continue
            elif cmp.is_satisfied_by(t):
              # t subsumes cmp
              break
            elif t.is_satisfied_by(cmp):
              targets.remove(t)
              targets.append(cmp.at(trans))
              # cmp replaces t (keep checking overlap)
            else:
              return False
    return True
