'''
component.py

Code that defines an abstract component type and related functions.

Copyright 2010 Peter Mawhorter
This software is distributed under the terms of the GNU General Public License.
'''

class ComponentType:
  '''
  A ComponentType is a type for components. The component types form one or
  more hierarchies, which are used to determine whether components are allowed
  to overlap. Each type has a name, some overlap planes, and a set of parents.
  '''

  name = "<unnamed>"
  planes = set()
  parents = set()

  def __init__(self, name, planes=None, parents=None):
    self.name = name
    self.planes = planes
    self.parents = parents or set()
  def __str__(self):
    return "component type '{}'".format(self.name)
  def __repr__(self):
    return "ComponentType({})".format(repr(self.name))

  def ancestors(self):
    '''
    Yields the (possibly over-)complete list of ancestors of this type, in
    arbitrary order. This type is not included in the list generated. Note that
    ancestors with multiple paths of inheritance will be generated multiple
    times.
    '''
    for p in self.parents:
      for a in p.ancestors():
        yield a
      yield p

  def is_subtype_of(self, other):
    '''
    Tests whether this type is a subtype of the given type.
    '''
    return (
      other == self
    or
      other in self.parents
    or
      any(p.is_strict_subtype_of(other) for p in self.parents)
    )

  def is_strict_subtype_of(self, other):
    '''
    Like is_subtype_of, but doesn't return true for self (i.e.
    MyType.is_subtype_of(MyType) is True, but
    MyType.is_strict_subtype_of(MyType) is False.)
    '''
    return (
      other in self.parents
    or
      any(p.is_strict_subtype_of(other) for p in self.parents)
    )

  def can_overlap(self, other):
    '''
    Checks whether this type can overlap with a component of the given type.
    Two components are allowed to overlap as long as don't share any overlap
    planes. Components that don't have an overlap plane set of None aren't
    allowed to overlap any other components.
    '''
    return self.planes != None\
      and other.planes != None\
      and not (self.planes & other.planes)

  def at(self, position):
    '''
    Returns a Component that's an instance of this ComponentType at the given
    position.
    '''
    return Component(self, position)

unknown = ComponentType("unknown")

class Component:
  '''
  A component is a minimal chunk of content. It has a real type, a required
  type, and a position. The real type designates the content that it
  represents, and together with the position specifies a concrete piece of
  content. The required type, on the other hand, specifies a more general
  requirement for the content. When fitting content together, components may be
  replaced with other components as long as their replacements conform to their
  required types. So for example, a component with type "powerup block" and
  required type "block" might be replaced by another component with type "coin
  block" and required type "platform". Note that the required type is ignored
  during replacement, but requirement generality escalation usually doesn't
  occur because new things are replaced with old things, rather than the other
  way around.
  '''

  real_type = unknown
  required_type = unknown
  position = None

  def __init__(self, type, position, required=None):
    self.real_type = type
    self.required_type = required or type
    self.position = position

  def __str__(self):
    return "a {} component at {}".format(self.real_type, self.position)

  def __repr__(self):
    if self.real_type == self.required_type:
      return "Component({}, {})".format(
        repr(self.real_type),
        repr(self.position),
      )
    else:
      return "Component({}, {}, {})".format(
        repr(self.real_type),
        repr(self.position),
        repr(self.required_type),
      )

  def is_instance_of(self, type):
    '''
    Tests whether this component is an instance of the given type. Uses this
    components real type, rather than its minimum required type.
    '''
    return self.real_type.is_subtype_of(type)

  def is_satisfied_by(self, other):
    '''
    Checks whether this component could be replaced by the given one. Doesn't
    check positions, just types. Note that this check doesn't use the
    required_type of the given component.
    '''
    return other.is_instance_of(self.required_type)

  def can_overlap(self, other):
    '''
    Checks whether this component can overlap the given one. Doesn't check
    positions, just types and overlap specifications.
    '''
    return self.real_type.can_overlap(other.real_type)

  def at(self, pos):
    '''
    Returns a new component based on this one at the given position.
    '''
    return Component(self.real_type, pos, self.required_type)
