'''
state.py

Contains the core algorithm code. Given a library of chunks to work with, it
produces a super-chunk by combining them using their anchor points.

Copyright 2010 Peter Mawhorter
This software is distributed under the terms of the GNU General Public License.
'''

import collections, random

from . import chunk

def remember_nothing(anchor, chunk, anchored_at, state, memories):
  '''
  A memory creation function that doesn't record anything.
  '''
  pass

def memories_of_boredom():
  '''
  A memory initialization function that creates a 'boredom' dictionary for use
  with the 'remember_boredom' memory creation function.
  '''
  return { "boredom": collections.defaultdict(lambda: 0) }

def remember_boredom(anchor, chunk, anchored_at, state, memories):
  '''
  A memory creation function that keeps track of how many times each anchor and
  chunk has been used in a 'boredom' dictionary.
  '''
  memories["boredom"][anchor] += 1
  memories["boredom"][chunk] += 1

def weighted_permutation(items, metric=lambda _: 1):
  '''
  Yields the given items permuted randomly according to their weights (as
  computed using the given metric function).
  '''
  pool = []
  total = 0
  for it in items:
    weight = metric(it)
    if weight > 0:
      pool.append([it, weight])
      total += weight
  #last = len(pool) + 1
  while pool:
    total = sum(w for it, w in pool)
    #print("wp", len(pool))
    #if len(pool) == last:
    #  raise ValueError(
    #    "Pool didn't diminish in size: {}, {}, {}".format(
    #      last, len(pool), total
    #    )
    #  )
    #last = len(pool)
    num = random.uniform(0, total)
    for i, (it, w) in enumerate(pool):
      num -= w
      if num <= 0:
        yield it
        pool.pop(i)
        break

class State:
  '''
  A state stores the state for the ORE algorithm. This includes the result so
  far, the library of chunks, metrics for anchor and chunk selection, and an
  arbitrary dictionary of memories, along with functions for initializing and
  recording memories. The methods of this class form the core of the algorithm.
  The anchor selection function should take an anchor and a state and return a
  weight for that anchor. The chunk selection function should take a chunk, an
  anchor within that chunk at which to attach it, and a state, and return a
  weight for that anchor. The current_anchor field of the state passed to the
  chunk selection function will store the anchor in the result at which the
  chunk is being considered for attachment. Note that the chunk selection
  function doesn't have to perform compatibility checking, since
  is_compatible_with is already called before a chunk can be chosen.
  '''
  library=None
  result=None
  anchor_metric=None
  chunk_metric=None
  memories=None
  memory_initializer=None
  memory_creator=None

  current_anchor=None

  def __init__(
    self,
    library,
    initial=None,
    anchor_metric=None,
    chunk_metric=None,
    memory_initializer=None,
    memory_creator=None,
  ):
    self.library = library
    self.result = initial or chunk.Chunk(anchors={Position.origin})
    self.exhausted = set()
    self.anchor_metric = anchor_metric or (lambda anchor, state: 1)
    self.chunk_metric = chunk_metric or (lambda chunk, anchored_at, state: 1)
    self.memory_initializer = memory_initializer or (lambda: {})
    self.memory_creator = memory_creator or remember_nothing
    self.memories = self.memory_initializer()
    self.current_anchor = None

  def reset_memory(self):
    '''
    Resets the algorithm memory using the memory initialization function.
    '''
    self.memories = self.memory_initializer()

  def step(self):
    '''
    Advances the algorithm one step by adding a single chunk from the library
    to the result. If it succeeds, it returns True, but if there isn't any way
    to add a chunk to the result, it returns False.
    '''
    #print("step")
    for anchor in self.ordered_anchors():
      self.current_anchor = anchor
      #print("anchor")
      for chunk, pos in self.ordered_chunks(anchor):
        #print("chunk found!")
        self.memory_creator(anchor, chunk, pos, self, self.memories)
        #print("at anchor:", self.result.get(anchor))
        #print("incoming:", chunk.get(pos))
        self.result.add_chunk(chunk, pos, anchor)
        #print("afterwards:", self.result.get(anchor))
        self.current_anchor = None
        return True
      self.exhausted.add(anchor)
      #print("no match")
    #print("fail")
    # Uh-oh... we're completely out of options.
    self.current_anchor = None
    return False

  def converge(self, condition, check_every=1):
    '''
    Repeatedly calls the step method, adding chunks to the result, until the
    result satisfies the given convergence condition (which should be a
    function that takes a State and returns a boolean). If 'check_every' is
    given, it specifies how many steps should be taken between checks (the
    default is 1, for checking the condition after every step). If the
    condition is already satisfied by the result, no steps are performed. This
    function returns True if it converges, or False if it gets stuck.
    '''
    while not condition(self):
      for i in range(check_every):
        if not self.step():
          return False
    return True

  def ordered_anchors(self):
    '''
    Uses the given anchor metric to select a position within the result at
    which to add a new chunk.
    '''
    if not self.result.anchors:
      raise ValueError(
        "Can't order anchors: no anchors exist in the result."
      )
    return weighted_permutation(
      self.result.anchors,
      (lambda anchor: self.anchor_metric(anchor, self))
    )

  def ordered_chunks(self, anchor):
    '''
    Uses the given chunk metric to order chunks from the library to add to
    the result at the given anchor. It yields the ordered chunks and the
    anchors within each chunk using which they should be aligned. If it is
    unable to find any suitable chunks for the selected anchor, it results in
    an empty list.
    '''
    ordered = weighted_permutation(
      self.library.anchored_chunks,
      (lambda c_and_a: self.chunk_metric(c_and_a[0], c_and_a[1], self))
    )
    #print("permutation")
    for chunk, anchored_at in ordered:
      #print("check")
      if (self.result.is_compatible_with(chunk, anchored_at, anchor)):
        #print("yes")
        yield (chunk, anchored_at)
      #print("no")
