'''
curate.py

Curates the Wesnoth data, working with the values defined in hierarchy.py.

Copyright 2010 Peter Mawhorter
This software is distributed under the terms of the GNU General Public License.
'''

import ore.chunk
import ore.position

import hierarchy

CELL_WIDTH = 12

def load(filename):
  with open(filename, 'rU') as fin:
    data = fin.read()
  chunks = data.split("Chunk:\n")[1:]
  lib = []
  for c in chunks:
    lib.append(unpack(c))
  return lib

def save(chunks, filename):
  with open(filename, 'w') as fout:
    for c in chunks:
      fout.write("Chunk:\n" + pack(c))

def terrain_string(components):
  '''
  Given a set of components, returns the terrain string for a hex containing
  them all. If for some reason there are more than there should be (for
  example, there should never be two different overlay components in a hex),
  some will be discarded at random.
  '''
  prefix = ''
  base = ''
  overlay = ''
  for c in components:
    if c.is_instance_of(hierarchy.starting_position):
      prefix = str(hierarchy.sp_to_nr[c.real_type]) + ' '
    else:
      s = hierarchy.type_to_char[c.real_type]
      if s[0] == '^':
        overlay = s[1:]
      elif '^' in s:
        base, overlay = s.split('^')
      else:
        base = s
  if overlay:
    overlay = '^' + overlay
  return (prefix + base + overlay).ljust(CELL_WIDTH)

def pack_terrain(chunk, height=None, width=None):
  '''
  Takes a chunk and returns a string that represents the terrain portion of
  that chunk, ignoring tags and anchors. If height and/or width are given,
  the chunk is trimmed or extended to match the given values.
  '''
  result = ''
  _width = 0
  _height = 0
  for pos in chunk.coverage:
    if pos.x >= _width:
      _width = pos.x + 1
    if pos.y >= _height:
      _height = pos.y + 1
  if height:
    _height = height
  if width:
    _width = width
  for y in range(_height):
    for x in range(_width):
      pos = ore.position.GridPosition(x, y)
      components = chunk.get(pos)
      if components:
        s = terrain_string(components)
      else:
        s = hierarchy.type_to_char[hierarchy.void].ljust(CELL_WIDTH)
      result += s + ','
    result = result[:-1] + '\n'
  return result

def build_map(chunk, filename, height=None, width=None):
  '''
  Takes a chunk and saves it as a map file. If height and/or width are given,
  they are used instead of the maximal component position from the given chunk
  (and any components outside of the bounds are cut off).
  '''
  raw = '''\
usage=map
border_size=1

'''
  raw += pack_terrain(chunk, height, width)
  with open(filename, 'w') as fout:
    fout.write(raw)

def pack(chunk):
  '''
  Takes a chunk and returns a string representing it (including its tags and
  anchors).
  '''
  result = 'Tags:\n'
  for t, v in chunk.tags.items():
    if v == True:
      result += t + '\n'
    else:
      result += "{} {}\n".format(t, v)

  result += 'Anchors:\n'
  for a in chunk.anchors:
    result += "{}, {}\n".format(a.x, a.y)

  result += 'Map:\n'
  result += pack_terrain(chunk)

  result += 'End\n'

  return result

def unpack(data):
  '''
  Takes a string representing a chunk (see the pack function) and returns the
  chunk.
  '''
  pos = None
  mode = None
  onex = ore.position.GridPosition(1, 0)
  offset = ore.position.GridPosition(0, 0)
  result = ore.chunk.Chunk()
  for line in data.split('\n'):
    if line == 'Tags:':
      mode = 'tags'
      continue
    if line == 'Offset:':
      mode = 'offset'
      continue
    if line == 'Anchors:':
      mode = 'anchors'
      continue
    if line == 'Map:':
      mode = 'map'
      pos = ore.position.GridPosition(0, 0)
      continue
    elif line == 'End':
      mode = None
      continue
    else:
      if mode == 'tags':
        things = line.strip().split()
        if len(things) == 1:
          result.tag(things[0], True)
        elif len(things) == 2:
          tag = things[0]
          val = things[1]
          try:
            val = int(val)
          except:
            try:
              val = float(val)
            except:
              pass
          result.tag(tag, val)
        else:
          raise ValueError("Tags can't have more than a name and a value.")
      if mode == 'offset':
        first, second = line.split(',')
        offset = ore.position.GridPosition(
          int(first.strip()),
          int(second.strip())
        )
      elif mode == 'anchors':
        first, second = line.split(',')
        anchor = ore.position.GridPosition(
          int(first.strip()),
          int(second.strip())
        )
        result.add_anchor(anchor + offset)
      elif mode == 'map':
        for s in line.split(','):
          optional = False
          base_required = None
          overlay_required = None
          s = s.strip()
          if '*' in s:
            s = s.replace('*', '')
            result.add_anchor(pos + offset)
          if '%' in s:
            s = s.replace('%', '')
            optional = True
          if ' ' in s:
            f, s = s.split()
            result.add_component(hierarchy.nr_to_sp[int(f)].at(pos + offset))
          if '|' in s:
            if '^' in s:
              b, o = s.split('^')
              if '|' in b:
                b, b_opt = b.split('|')
                if b_opt in hierarchy.char_to_type:
                  base_required = hierarchy.char_to_type[b_opt]
                else:
                  raise ValueError("Bad required base type: '{}'".format(b_opt))
              if '|' in o:
                o, o_opt = o.split('|')
                o_opt = '^' + o_opt
                if o_opt in hierarchy.char_to_type:
                  overlay_required = hierarchy.char_to_type[o_opt]
                else:
                  raise ValueError(
                    "Bad required overlay type: '{}'".format(o_opt)
                  )
              s = b + '^' + o
            else:
              s, opt = s.split('|')
              if opt in hierarchy.char_to_type:
                base_required = hierarchy.char_to_type[opt]
          if s in hierarchy.char_to_type:
            cmp = hierarchy.char_to_type[s].at(pos + offset)
            if base_required:
              cmp = hierarchy.with_required(cmp, base_required)
            elif optional:
              cmp = hierarchy.optional(cmp)
            result.add_component(cmp)
          else:
            try:
              b, o = s.split('^')
              base = hierarchy.char_to_type[b].at(pos + offset)
              overlay = hierarchy.char_to_type['^' + o].at(pos + offset)
              if optional:
                base = hierarchy.optional(base)
                overlay = hierarchy.optional(overlay)
              if base_required:
                base = hierarchy.with_required(base, base_required)
              if overlay_required:
                overlay = hierarchy.with_required(overlay, overlay_required)
              result.add_component(base)
              result.add_component(overlay)
            except:
              if s.strip():
                raise ValueError("Bad map cell: '{}'".format(s))
          pos += onex
        pos = ore.position.GridPosition(0, pos.y + 1)
  return result
