'''
metrics.py

The metrics that constrain and control generation via ORE.

Copyright 2010 Peter Mawhorter
This software is distributed under the terms of the GNU General Public License.
'''

import collections

import ore.chunk
import ore.state
import ore.position

import hierarchy

def init_memory():
  return {
    "boredom": collections.defaultdict(lambda: 0),
    "villages_so_far": 0
  }

def remember_villages(anchor, chunk, anchored_at, state, memories):
  #print("placed:", chunk.tags, anchor, chunk)
  if "village" in chunk.tags:
    memories["villages_so_far"] += 1

def record_memory(*args):
  ore.state.remember_boredom(*args)
  remember_villages(*args)
  #print("alter?", alters(args[1], args[2], args[3]))

def positions_near(position, distance, chunk, filter):
  result = 0
  for x in range(position.x - distance, position.x + distance + 1):
    for y in range(position.y - distance, position.y + distance + 1):
      components = chunk.get(ore.position.GridPosition(x, y))
      if filter(components):
        result += 1
  return result

def build_anchor_metric(width, height):
  '''
  Returns an anchor metric tuned to the given level width and height. The
  returned metric computes a weight for a given anchor. It selects anchors near
  unfilled space, and won't select anchors that are outside the level.
  '''
  def anchor_metric(anchor, state):
    '''
    An anchor metric with a given width and height that excludes anchors
    outside of the level and which selects anchors near unfilled space.
    '''
    if not (0 <= anchor.x < width) or not (0 <= anchor.y < height):
      return 0
    return positions_near(
      anchor,
      2,
      state.result,
      lambda components: all(
        "base" not in c.real_type.planes for c in components
      )
    )
  return anchor_metric

def flavors(chunk, position):
  result = set()
  for c in chunk.get(position):
    result.update(hierarchy.movement_categories_of[c.real_type])
  return result

def alters(chunk, anchored_at, state):
  for p in chunk.coverage:
    loc = state.current_anchor + p - anchored_at
    contents = { c.real_type for c in state.result.get(loc) }
    compare = { c.real_type for c in chunk.get(p) }
    if contents != compare:
      return True
  return False

def build_chunk_metric(width, height):
  '''
  Returns a chunk metric tuned to the given level width and height. The
  returned metric computes a weight for a given chunk.
  '''
  def chunk_metric(chunk, anchored_at, state):
    '''
    Returns a weight for the given chunk, anchored at the given location, being
    inserted into the given state.
    '''
    # Disqualifications:
    if "disabled" in chunk.tags:
      return 0
  
    odd = (state.current_anchor.x - anchored_at.x) % 2
    if "odd_column" in chunk.tags and not odd:
      return 0
    elif "even_column" in chunk.tags and odd:
      return 0
  
    current_flavors = flavors(state.result, state.current_anchor)
    proposed_flavors = flavors(chunk, anchored_at)
    if current_flavors and proposed_flavors\
    and not (current_flavors & proposed_flavors):
      return 0
  
    times_used = state.memories["boredom"][chunk]
    if "unique" in chunk.tags and times_used > 0:
      return 0
  
    if not alters(chunk, anchored_at, state):
      return 0

    if "edge-shy" in chunk.tags:
      if not (
        3 < state.current_anchor.x < width - 3
        and 3 < state.current_anchor.y
      ):
        return 0;
  
    # The weight value:
    weight = 0.94 ** times_used
  
    # Weight adjustments:
    if "transition" in chunk.tags:
      weight *= 1000 * 0.001 ** times_used
    if "structured" in chunk.tags and times_used > 0:
      weight *= 0.01 ** times_used
    if "village" in chunk.tags:
      #print("V!")
      vm = 1
      villages_so_far = state.memories["villages_so_far"]
      if villages_so_far > 5:
        vm *= 0.1 ** villages_so_far
      else:
        vm *= 5 ** (5 - villages_so_far)
        vm *= 50
      nearby = positions_near(
        state.current_anchor,
        5,
        state.result,
        lambda components: any(
          c.is_instance_of(hierarchy.village) for c in components
        )
      )
      vm *= 0.0001 ** nearby
      weight *= vm
      #print("v:", villages_so_far, vm)
    if "filler" in chunk.tags:
      weight *= 0.0001
    if "road" in chunk.tags:
      weight *= 0.000000005
    if "rare" in chunk.tags:
      weight *= 0.0001
    if "sparse" in chunk.tags:
      weight *= 0.005 ** times_used
    return weight
  return chunk_metric

def build_completion_metric(width, height):
  '''
  Builds a completion metric for the given width and height. The metric
  constructed returns True when the entire level is filled.
  '''
  def completion_metric(state):
    '''
    The convergence condition: the level is declared done as soon as there are
    no more empty spaces left.
    '''
    chunk = state.result
    for x in range(width):
      for y in range(height):
        components = chunk.get(ore.position.GridPosition(x, y))
        has_base = False
        for c in components:
          if "base" in c.real_type.planes:
            has_base = True
        if not has_base:
          return False
    return True
  return completion_metric
